@@ -623,7 +623,12 @@ static const char *cleanup_patch_header(CLEANUP_STATE *state,
 	    msg_warn("%s: seek file %s: %m", myname, cleanup_path);
 	    CLEANUP_PATCH_HEADER_RETURN(cleanup_milter_error(state, errno));
 	}
-	CLEANUP_OUT_BUF(state, rec_type, buf);
+	/* The saved "append header" pointer record may still contain "0". */
+	if (saved_read_offset == state->append_hdr_pt_offset)
+	    cleanup_out_format(state, REC_TYPE_PTR, REC_TYPE_PTR_FORMAT,
+			       (long) state->append_hdr_pt_target);
+	else
+	    CLEANUP_OUT_BUF(state, rec_type, buf);
 	if (msg_verbose > 1)
 	    msg_info("%s: %ld: write %.*s", myname, (long) write_offset,
 		     LEN(buf) > 30 ? 30 : (int) LEN(buf), STR(buf));
@@ -779,6 +784,10 @@ static const char *cleanup_upd_header(void *context, ssize_t index,
      * The lookup result will never be a pointer record.
      * 
      * Index 1 is the first matching header instance.
+     * 
+     * XXX When a header is updated repeatedly we create jumps to jumps. To
+     * eliminate this, rewrite the loop below so that we can start with the
+     * pointer record that points to the header that's being edited.
      */
 #define DONT_SAVE_RECORD	0
 #define NO_PTR_BACKUP		0
@@ -821,12 +830,16 @@ static const char *cleanup_upd_header(void *context, ssize_t index,
 		CLEANUP_UPD_HEADER_RETURN(cleanup_milter_error(state, errno));
 	    }
 	    if (rec_type == REC_TYPE_PTR) {
+		/* The "append header" pointer record content must be saved. */
+		if (saved_read_offset == state->append_hdr_pt_offset)
+		    break;
 		if (jumped == 0) {
 		    /* Enough contiguous space for writing a PTR record. */
 		    avail_space += read_offset - saved_read_offset;
 		    jumped = 1;
 		}
-		if (rec_goto(state->dst, STR(rec_buf)) < 0) {
+		if (rec_goto(state->dst, STR(rec_buf)) < 0
+		    || (read_offset = vstream_ftell(state->dst)) < 0) {
 		    msg_warn("%s: read file %s: %m", myname, cleanup_path);
 		    CLEANUP_UPD_HEADER_RETURN(cleanup_milter_error(state,
 								   errno));
@@ -1611,6 +1624,8 @@ static void usage(void)
     msg_warn("    del_rcpt addr");
 }
 
+/* flatten_args unparse partial command line */
+
 static void flatten_args(VSTRING *buf, char **argv)
 {
     char  **cpp;
@@ -1624,6 +1639,8 @@ static void flatten_args(VSTRING *buf, char **argv)
     VSTRING_TERMINATE(buf);
 }
 
+/* open_queue_file - open an unedited queue file (all-zero dummy PTRs) */
+
 static void open_queue_file(CLEANUP_STATE *state, const char *path)
 {
     VSTRING *buf = vstring_alloc(100);
@@ -1634,6 +1651,13 @@ static void open_queue_file(CLEANUP_STATE *state, const char *path)
     long    rcpt_count;
     long    qmgr_opts;
 
+    if (state->dst != 0) {
+	msg_warn("closing %s", cleanup_path);
+	vstream_fclose(state->dst);
+	state->dst = 0;
+	myfree(cleanup_path);
+	cleanup_path = 0;
+    }
     if ((state->dst = vstream_fopen(path, O_RDWR, 0)) == 0) {
 	msg_warn("open %s: %m", path);
     } else {
@@ -1761,7 +1785,7 @@ int     main(int unused_argc, char **argv)
 		msg_warn("bad add_header argument count: %d", argv->argc);
 	    } else {
 		flatten_args(arg_buf, argv->argv + 2);
-		cleanup_add_header(state, argv->argv[2], STR(arg_buf));
+		cleanup_add_header(state, argv->argv[1], STR(arg_buf));
 	    }
 	} else if (strcmp(argv->argv[0], "ins_header") == 0) {
 	    if (argv->argc < 3) {
